/*
 * Copyright (C) 2007 Richard Hughes <richard@hughsie.com>
 *
 * Licensed under the GNU Lesser General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <sys/stat.h>
#include <glib/gstdio.h>
#include <glib/gprintf.h>

#include <gmodule.h>
#include <glib.h>
#include <ohm-plugin.h>
#include <dbus/dbus-glib.h>

#define	HAL_DBUS_SERVICE		 	"org.freedesktop.Hal"
#define	HAL_DBUS_INTERFACE_POWER	 	"org.freedesktop.Hal.Device.SystemPowerManagement"
#define HAL_ROOT_COMPUTER		 	"/org/freedesktop/Hal/devices/computer"

enum {
	CONF_BUTTON_POWER_CHANGED,
	CONF_BUTTON_LID_CHANGED,
	CONF_TIMEOUTS_POWERDOWN_CHANGED,
	CONF_INHIBIT_CHANGED,
	CONF_LAST
};

static gboolean
do_suspend(OhmPlugin *plugin)
{
	FILE *fp;

	g_debug("in %s", G_STRFUNC);
	fp = g_fopen("/sys/power/state", "w");
	if (!fp)
		g_warning ("Couldn't open /sys/power/state.");

	if (g_fprintf(fp, "mem") < 0)
		g_warning ("Couldn't write to /sys/power/state.");

	fclose(fp);

	g_debug("should have suspended");
	return TRUE;
}

static gboolean
pre_sleep(OhmPlugin *plugin)
{
  	g_debug("in %s", G_STRFUNC);
	
	/* Turn off the backlight when sleeping. */ 
	ohm_plugin_conf_set_key (plugin, "backlight.state", 0);

	return TRUE;
}

static gboolean
pre_suspend(OhmPlugin *plugin)
{
  	g_debug("in %s", G_STRFUNC);
        return TRUE;
}

static gboolean
post_suspend(OhmPlugin *plugin)
{
  	g_debug("in %s", G_STRFUNC);
	
	/* We've resumed now.  Bring the backlight back immediately. */
	ohm_plugin_conf_set_key (plugin, "backlight.state", 1);
        
	return TRUE;
}

static gboolean
post_sleep(OhmPlugin *plugin)
{
	char source[256];
	FILE *fp;

	g_debug("in %s", G_STRFUNC);
	
        fp = g_fopen("/sys/power/wakeup-source", "r");
        if (!fp)
                g_warning ("Couldn't open /sys/power/wakeup-source.");

	while (fgets(source, sizeof(source), fp) != NULL) {
		g_debug ("Source was %s.", source);
	}

	if (g_ascii_strncasecmp(source, "power button", 12) == 0) {
		/* Wake up fully. */
		g_debug("Waking up fully.");	  
		ohm_plugin_conf_set_key (plugin, "backlight.state", 1);
	}
	else {
		/* Go back to sleep. */
		g_debug("Going back to sleep.");
		do_suspend(plugin);
		post_sleep(plugin);
	}
	
	return TRUE;
}

static void
power_button_pressed (OhmPlugin *plugin)
{
	gint value;

	/* check system inhibit - this is broken as any client can unref all */
	ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &value);
	if (value == 1) {
		g_warning ("not doing lid closed action as inhibited");
		return;
	}

	pre_sleep(plugin);
	do_suspend(plugin);
	g_debug("Back in power_button_pressed.");
	post_sleep(plugin);
}

static void
lid_closed (OhmPlugin *plugin, gboolean is_closed)
{
	gint value;

	if (is_closed == FALSE) {
		return;
	}

	/* check system inhibit - this is broken as any client can unref all */
	ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &value);
	if (value == 1) {
		g_warning ("not doing lid closed action as inhibited");
		return;
	}

	pre_sleep(plugin);
	do_suspend(plugin);
	post_sleep(plugin);
}

static void
system_is_idle (OhmPlugin *plugin)
{
#define SUSPEND_TAG_FILE "/etc/ohm/inhibit-suspend"
  
	gint value;
	gint ac;
	struct stat stat_buf;

	g_debug("in %s", G_STRFUNC);

	/* check for inhibit file */
	if (stat(SUSPEND_TAG_FILE, &stat_buf) >= 0) {
		/* it exists */
		g_debug("not suspending; inhibit file found.");
		return;
	}
	
	/* check system inhibit - this is broken as any client can unref all */
	ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &value);

	if (value == 1) {
		g_debug ("not doing idle action as inhibited");
		return;
	}

	/* check AC state */
	ohm_plugin_conf_get_key (plugin, "acadapter.state", &ac);

	if (ac == 1) {
	        g_debug ("not suspending while on AC");
		return;
	}

	pre_suspend(plugin);
	do_suspend(plugin);
	post_suspend(plugin);
}

/**
 * plugin_notify:
 * @plugin: This class instance
 *
 * Notify the plugin that a key marked with ohm_plugin_conf_interested ()
 * has it's value changed.
 * An enumerated numeric id rather than the key is returned for processing speed.
 */
static void
plugin_notify (OhmPlugin *plugin, gint id, gint value)
{
	if (id == CONF_BUTTON_LID_CHANGED) {
		if (value == 1) {
			lid_closed (plugin, TRUE);
		} else {
			lid_closed (plugin, FALSE);
		}
	} else if (id == CONF_BUTTON_POWER_CHANGED) {
		/* only match on button press, not release */
		if (value == 1) {
			power_button_pressed (plugin);
		}
	} else if (id == CONF_TIMEOUTS_POWERDOWN_CHANGED) {
		/* only match on idle, not reset */
		if (value == 1) {
			g_debug ("powerdown is 1, suspending");
			system_is_idle (plugin);
		}
	}
}

OHM_PLUGIN_DESCRIPTION (
	"suspend",			/* description */
	"0.0.1",			/* version */
	"richard@hughsie.com",		/* author */
	OHM_LICENSE_LGPL,		/* license */
	NULL,				/* initialize */
	NULL,				/* destroy */
	plugin_notify			/* notify */
);

OHM_PLUGIN_SUGGESTS (
	"timeouts",
	"buttons"
);

OHM_PLUGIN_INTERESTED (
	{"button.power", CONF_BUTTON_POWER_CHANGED},
	{"button.lid", CONF_BUTTON_LID_CHANGED},
	{"timeouts.powerdown", CONF_TIMEOUTS_POWERDOWN_CHANGED},
	{"suspend.fixme_inhibit", CONF_INHIBIT_CHANGED}
);
